--- arr2
-- @author Enno Sylvester
-- @copyright 2019, 2021
-- @module eatglobal.arr2

local historie = [[
	Version 1.0:
	 - Release
	Version 1.1:
	 - Added: Added Fields "__debug__" and "historie" to result-table arr2.
	 - Changed: arr2.new now accepts arr2, arr3, vec2, vec3, vec4 and numbers as parameters.
	 - Changed: +, -, * and / now accept both arr2 and numbers.
	 - Changed: Functions with table returns aka lists, returns now as eatSetEx.
	 - Changed: Deletion of all import functions, because arr2.new covers all possibilities.
	 - Changed: Deletion of the functions arr2.copy and arr2.mt, because arr2.new covers these functionalities. Class function copy is retained.
	Version 1.2:
	 - Added: Modulfähigkeit (Erweiterungsmöglichkeiten durch Module) implementiert.
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 2,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local vec2
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		vec2 = global.vec2
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local function getArr2(x, y)
	if (type(x) == "table") then
		if ((type(x[1]) == "number") and (type(x[2]) == "number")) then
			-- x is a arr2 or arr3
			return {x[1], x[2]}
		elseif ((type(x.x) == "number") and (type(x.y) == "number")) then
			-- x is a vec2 (or maybe a vec3 or vec4)
			return {x.x, x.y}
		end
	elseif (type(x) == "number") then
		-- x is a number; We look to see if y are also a number. Otherwise they get x as value
		return {x, (type(y) == "number") and y or x}
	end
	-- nothing; return default
	return {0, 0}
end

local function calcPointData(p, radius)
	if (radius == 0) then
		-- Daten
		return {
			aRad = 0,
			yLength = p[2],
			tangent = functionClass.new(0, 1),
		}
	else
		local r = radius + p[1]
		local aRad = p[2] / radius
		local yLength = aRad * r
		-- Daten
		return {
			aRad = aRad,
			yLength = yLength,
			tangent = functionClass.new(-math.sin(aRad), math.cos(aRad), 0),
		}
	end
end

local mt
mt = {
  __add = function (self, rhs) -- "add" event handler
		if (type(rhs) == "number") then
			return setmetatable({self[1] + rhs, self[2] + rhs}, mt)
		else
			local v = getArr2(x, y)
			return setmetatable({self[1] + v[1], self[2] + v[2]}, mt)
		end
  end,
	__sub = function (self, rhs) -- "sub" event handler
		if (type(rhs) == "number") then
			return setmetatable({self[1] - rhs, self[2] - rhs}, mt)
		else
			local v = getArr2(x, y)
			return setmetatable({self[1] - v[1], self[2] - v[2]}, mt)
		end
  end,
	__mul = function (self, rhs) -- "mul" event handler
		if (type(rhs) == "number") then
			return setmetatable({self[1] * rhs, self[2] * rhs}, mt)
		else
			local v = getArr2(x, y)
			return setmetatable({self[1] * v[1], self[2] * v[2]}, mt)
		end
	end,
	__div = function (self, rhs) -- "div" event handler
		if (type(rhs) == "number") then
			return setmetatable({self[1] / rhs, self[2] / rhs}, mt)
		else
			local v = getArr2(x, y)
			return setmetatable({self[1] / v[1], self[2] / v[2]}, mt)
		end
	end,
	__index = {
		copy = function(self)
			return setmetatable({self:toParams()}, mt)
		end,
		degToRad = function(self)
			return setmetatable({math.rad(self[1]), math.rad(self[2])}, mt)
		end,
		distance = function(self, p)
			return (self - p):length()
		end,
		getMiddle = function(self, p)
			return self + ((functionClass.new(p) - self) / 2)
		end,
		length = function(self)
			return math.sqrt(self[1]^2 + self[2]^2)
		end,
		lengthTo = function(self, p, radius)
			local r = radius or 0
			if (r == 0) then
				return self:distance(p)
			else
				local d0 = calcPointData(self, radius)
				local d1 = calcPointData(p, radius)
				return d1.yLength - d0.yLength
			end
		end,
		swap = function(self)
			return setmetatable({self[2], self[1]}, mt)
		end,
		toParams = function(self)
			return self[1], self[2]
		end,
		toString = function(self)
			return tostring(self)
		end,
		toVec = function(self)
			return vec2.new(self)
		end,
	},
	__tostring = function(self)
		return string.format("{%.4f, %.4f}", self:toParams())
	end,
}

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	new
	--	
	------------------------------------------------------------------------------------
	new = function(x, y)
		return setmetatable(getArr2(x, y), mt)
	end,
	------------------------------------------------------------------------------------
	--	mtPoints
	--	
	------------------------------------------------------------------------------------
	mtPoints = function(points)
		local result = eatSetEx.new()
		for i = 1, #points do
			result:insert(functionClass.new(points[i]))
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	length
	--	
	------------------------------------------------------------------------------------
	length = function(p)
		return functionClass.new(p):length()
	end,
	------------------------------------------------------------------------------------
	--	distance
	--	
	------------------------------------------------------------------------------------
	distance = function(p0, p1)
		return functionClass.new(p0):distance(p1)
	end,
	------------------------------------------------------------------------------------
	--	lengthBetween
	--	
	------------------------------------------------------------------------------------
	lengthBetween = function(p0, p1, radius)
		return functionClass.new(p0):lengthTo(p1, radius)
	end,
	------------------------------------------------------------------------------------
	--	toVec
	--	
	------------------------------------------------------------------------------------
	-- Exportfunktionen
	toVec = function(p)
		return vec2.new(p)
	end,
	------------------------------------------------------------------------------------
	--	toParams
	--	
	------------------------------------------------------------------------------------
	toParams = function(p)
		return p[1], p[2]
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
