--- fileutils
-- @author Enno Sylvester
-- @copyright 2022, 2023
-- @module eatglobal.fileutils

local historie = [[
	Version 1.1:
	 - Release
	Version 1.2:
	 - Changed:	function mapFiles - Optionales (Parameter subDirs) Scannen von Unterverzeichnissen implementiert.
	Version 1.3:
	 - Added: function openFile, function getTempFilename
	 - Changed:	function openTempFile überarbeitet.
]]

------------------------------------------------------------------------------------
--	Aktuelle Versionsdaten.
------------------------------------------------------------------------------------
local versionsData = {
	majorVersion = 1,
	minorVersion = 3,
}

------------------------------------------------------------------------------------
--	isWindows
------------------------------------------------------------------------------------
local isWindows = (os.getenv("OS") == "Windows_NT")

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local math
local string
local table
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		math = global.math
		string = global.string
		table = global.table
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
local defaultDebugLevel = 2
local minDebugLevel = 1
local maxDebugLevel = 5

local function quoteParam(fileName)
	return '"'..string.gsub(fileName, '/', '\\')..'"'
end

functionClass = {
	------------------------------------------------------------------------------------
	--	getFileName
	--	Extrahiert den Dateinamen (incl. Extension) aus 'fileName'.
	--
	--	"c:/werner/werner.txt"	-->	"werner.txt"
	--	"c:/werner/.txt"				-->	".txt"
	--	"c:/werner/."						-->	"."
	--	"c:/werner/werner"			-->	"werner"
	--	"c:/werner/"						-->	""
	--	"c:/"										-->	""
	------------------------------------------------------------------------------------
	getFileName = function(fileName)
		if global.misc.isValidId(fileName) then
			local sLen = fileName:len()
			for i = sLen, 1, -1 do
				if (i == 1) then
					return fileName
				else
					local char = fileName:sub(i, i)
					if ((char == "/") or (char == ":")) then
						return fileName:sub(i + 1, sLen)
					end
				end
			end
		end
		return ""
	end,

	------------------------------------------------------------------------------------
	--	getFileData
	--	Ermittelt Pfad, Dateiname  und Dateiendung aus 'fileName'.
	--	Rückgabe ist eine Tabelle mit den Einträgen 'path', 'name' und 'ext'.
	------------------------------------------------------------------------------------
	getFileData = function(fileName)
		local fName = functionClass.getFileName(fileName)
		local data = string.split(fName, ".")
		return {
			name = data[1] or "",
			ext = data[2] or "",
			path = (fName ~= "") and fileName:sub(1, fileName:find(fName, 1, true) - 1) or fileName,
		}
	end,

	------------------------------------------------------------------------------------
	--	getFileExtension
	--	Extrahiert die Dateiendung aus 'fileName'.
	--	Ist 'includeDot' == true, so wird der '.' mit ausgegeben (z.B. ".lua", ".txt").
	------------------------------------------------------------------------------------
	getFileExtension = function(fileName, includeDot)
		local data = functionClass.getFileData(fileName)
		return ((includeDot == true) and (data.ext ~= "")) and "."..data.ext or data.ext
	end,

	------------------------------------------------------------------------------------
	--	getFileNameAndExtension
	--	Gibt den Dateinamen ud die Dateierweiterung zurück.
	--	Ist 'includePath' == true wird der Dateiname inclusive Pfad zurückgegeben.
	--
	--	fileName = "mods/eat1963_testmod_2/res/scripts/www.lua"; includePath = false oder nil
	--	--> www			lua
	--	fileName = "mods/eat1963_testmod_2/res/scripts/www.lua"; includePath = true
	--	--> mods/eat1963_testmod_2/res/scripts/www			lua
	------------------------------------------------------------------------------------
	getFileNameAndExtension = function(fileName, includePath)
		local data = functionClass.getFileData(fileName)
		return (includePath ~= true) and data.name or data.path..data.name, data.ext
	end,

	------------------------------------------------------------------------------------
	--	fileNameIncludeSuffix
	--	Fügt 'suffix' zwischen dem Dateinamen und der Dateiendung ein.
	--	Bei 'includePath' == true Rückgabe incl. Pfad.
	--
	--	fileutils.fileNameIncludeSuffix("mods/eat1963_testmod_2/res/scripts/www.lua", "_001", true)
	--	--> mods/eat1963_testmod_2/res/scripts/www_001.lua
	------------------------------------------------------------------------------------
	fileNameIncludeSuffix = function(fileName, suffix, includePath)
		local fName, fExt = functionClass.getFileNameAndExtension(fileName, includePath)
		return global.misc.isValidId(fExt) and fName..suffix.."."..fExt or fName..suffix
	end,

	------------------------------------------------------------------------------------
	--	getModulFileName
	--	Ermittelt den Dateinamen des aufrufenden Moduls.
	--	Rückgabe incl. Pfad.
	------------------------------------------------------------------------------------
	getModulFileName = function(levelUp)
		local level = math.clamp(defaultDebugLevel + global.misc.getValue(levelUp, 0), minDebugLevel, maxDebugLevel)
		math.min(math.max(defaultDebugLevel + global.misc.getValue(levelUp, 0), minDebugLevel), maxDebugLevel)
		local info = debug.getinfo(level, "S")
		return (info ~= nil) and string.gsub(string.gsub(info.source, '\\', '/'), '@', "") or ""
	end,

	------------------------------------------------------------------------------------
	--	getModPath
	--	Ermittelt den Pfad des aufrufenden Moduls.
	------------------------------------------------------------------------------------
	getModPath = function(levelUp)
		local level = math.clamp(defaultDebugLevel + global.misc.getValue(levelUp, 0), minDebugLevel, maxDebugLevel)
		local info = debug.getinfo(level, "S")
		local modulFile = (info ~= nil) and string.gsub(string.gsub(info.source, '\\', '/'), '@', "") or ""
		local ix = modulFile:find("/res/")
		return (ix ~= nil) and modulFile:sub(1, ix) or functionClass.getFileData(modulFile).path
	end,

	------------------------------------------------------------------------------------
	--	getFileFolders
	--	Gibt ein Array mit allen Verzeichnissen von 'fileName' zurück.
	--	Typ der Rückgabe ist ein eatSetContainer.
	--
	--	fileName:	mods/eat1963_testmod_2/res/scripts/www.lua
	--	--> { "mods", "eat1963_testmod_2", "res", "scripts" }
	------------------------------------------------------------------------------------
	getFileFolders = function(fileName)
		--return string.split(functionClass.getFileData(fileName).path, "/"):filter(function(value) return global.misc.isValidId(value) end)
		return string.split(functionClass.getFileData(fileName).path, "/", nil, function(value) return global.misc.isValidId(value) end)
	end,

	------------------------------------------------------------------------------------
	--	fileExists
	--	Prüft, ob 'fileName' real vorhanden ist (Rückgabe true).
	------------------------------------------------------------------------------------
	fileExists = function(fileName)
		local file = io.open(fileName, "r")
		if file then
			file:close()
			return true
		end
		return false
	end,

	------------------------------------------------------------------------------------
	--	readData
	--	Liest Daten aus 'fileName' und gibt diese zurück.
	--	'dataFormat' kann "TEXT", "BLANK", "FUNC_DATA", "RETURN" sein.
	--	Ist 'dataFormat' ungültig wird "RETURN" gewählt.
	--	Ist 'dataFormat' <> "TEXT" bzw. "FUNC_DATA", 'resetFileOnError' == true und
	--	'fileName' enthält keine gültigen Daten, so wird 'fileName' mit einer leeren Tabelle gefüllt.
	------------------------------------------------------------------------------------
	readData = function(fileName, dataFormat, resetFileOnError)
		local function logError(err)
			logFnTraceback(string.format("[fileutils.readData] - %s", err))
		end
		
		local function readToEnv(env)
			setmetatable(env, {__index = _G})
			local file, err = loadfile(fileName, "tb", env)
			if (file~= nil) then
				file()
				return true
			else
				logError(string.format("Datei: '%s' | Fehler: '%s'", fileName, err))
				return false, err
			end
		end
		----------------------------------------
		dataFormat = global.misc.getValue(dataFormat, "RETURN")
		if (dataFormat == "FUNC_DATA") then
			local env = {
				data = function()
					return {}
				end
			}
			local result, err = readToEnv(env)
			if result then
				if (env.data ~= nil) then
					return env.data()
				else
					logError(string.format("Datei '%s' enthält keine Funktion 'data()'.", fileName))
					return {}
				end
			else
				return {}, err
			end
		elseif (dataFormat == "TEXT") then
			local file = io.open(fileName, "r")
			if file then
				local s = file:read("*a")
				file:close()
				return s
			else
				return "", string.format("Datei '%s' kann nicht geöffnet werden.", fileName)
			end
		else
			local file, err = loadfile(fileName)
			if (file ~= nil) then
				local result = file()
				if (type(result) ~= "table") then
					result = {}
					if (resetFileOnError == true) then
						functionClass.saveData({}, fileName)
					end
					logError(string.format("Datei '%s' ist keine gültige Datendatei. Evtl. ist sie beschädigt?!", fileName))
				end
				return result
			else
				if functionClass.fileExists(fileName) then
					logError(string.format("Datei: '%s' | Fehler: '%s'", fileName, err))
				end
				return {}, err
			end
		end
	end,

	------------------------------------------------------------------------------------
	--	saveData
	--	Speichert 'data' in 'fileName'.
	--	'dataFormat' kann "TEXT", "BLANK", "FUNC_DATA", "RETURN" sein.
	--	Ist 'dataFormat' ungültig wird "RETURN" gewählt.
	--	Ist 'dataFormat' == "TEXT" und 'append' == true, so wird die Datei um 'data'
	--	erweitert. Andernfalls wird die Datei gellert und mit 'data' gefüllt.
	------------------------------------------------------------------------------------
	saveData = function(data, fileName, dataFormat, append)
		dataFormat = global.misc.getValue(dataFormat, "RETURN")
		local file = io.open(fileName, ((dataFormat == "TEXT") and (append == true)) and "a" or "w")
		if file then
			if (dataFormat == "BLANC") then
				file:write(eatToString(data))
			elseif (dataFormat == "FUNC_DATA") then
				file:write("function data()\nreturn "..eatToString(data).."\nend\n")
			elseif (dataFormat == "TEXT") then
				file:write(data)
			else
				file:write("return "..eatToString(data).."\n")
			end
			file:close()
		end
	end,

	------------------------------------------------------------------------------------
	--	deleteFile
	--	Löscht 'fileName'.
	------------------------------------------------------------------------------------
	deleteFile = function(fileName)
		functionClass.deleteFiles(fileName)
	end,

	------------------------------------------------------------------------------------
	--	deleteFiles
	--	Löscht eine oder mehrere Dateien.
	--	files kann String mit einem Dateinamen oder Tabelle mit mehreren Dateinamen sein.
	------------------------------------------------------------------------------------
	deleteFiles = function(files)
		------------------------------------------------------------------------------------
		--	doIt
		------------------------------------------------------------------------------------
		local function doIt(fileTable)
			if isWindows then
				local tmpFile, tmpFileName = functionClass.openTempFile()
				if tmpFile then
					local exeFilename = global.apiPath.."eatglobal/bin/delfiles.cmd"
					local cmd = '"'..quoteParam(exeFilename)..' '..quoteParam(tmpFileName)..'"'
					for fileName in eatSetContainer(fileTable):_pairs() do
						tmpFile:write(quoteParam(fileName), "\n")
					end
					tmpFile:close()
					local p = io.popen(cmd)
					if p then
						p:close()
					end
				end
			else
				for fileName in eatSetContainer(fileTable):_pairs() do
					if functionClass.fileExists(fileName) then
						local cmd = '"rm '..quoteParam(fileName)..'" '
						local p = io.popen(cmd)
						if p then
							p:close()
						end
					end
				end
			end
		end
		
		------------------------------------------------------------------------------------
		------------------------------------------------------------------------------------
		if (type(files) == "string") then
			doIt({files})
		elseif (type(files) == "table") then
			doIt(files)
		end
	end,
	
	------------------------------------------------------------------------------------
	--	listDirs
	--	Listet alle Unterverzeichnisse von 'dir' auf.
	--	Ist 'includeSubDirs' == true, so werden auch alle Unter-, Unter-, ....
	--	Unterverzeichnisse ausgegeben.
	--	Rückgabe ist eatset.
	------------------------------------------------------------------------------------
	listDirs = function(dir, includeSubDirs)
		local function getSubDirs(subDir)
			local result = eatSetEx.new()
			local f = io.popen(string.format([[dir "%s" /b /ad]], subDir))
			if f then
				for s in f:lines() do
					if (s:len() > 0) then
						result:insert(string.format([[%s%s/]], subDir, s))
					end
				end
				f:close()
			end
			
			return result
		end
		
		local function doIt(subDir, result)
			local suDirs = getSubDirs(subDir)
			if (includeSubDirs == true) then
				for _dir in suDirs:_pairs() do
					result:insert(_dir)
					doIt(_dir, result)
				end
			else
				result:append(suDirs)
			end
			
			return result
		end
		---------------------------------------------------
		return doIt(dir, eatSetEx.new())
	end,

	------------------------------------------------------------------------------------
	--	mapFiles
	--	
	------------------------------------------------------------------------------------
	mapFiles2 = function(dirs, mapFn, filterFn)
		local doFilter = global.misc.buildFilterFn(filterFn)
		local result = eatSetEx.new()
		------------------------------------------------------------------------------------
		--	doIt
		------------------------------------------------------------------------------------
		local function doIt(dirsTable)
				local tmpFile, tmpFileName = functionClass.openTempFile()
				if tmpFile then
					local exeFilename = global.apiPath.."eatglobal/bin/listdirs"..(isWindows and ".cmd" or ".sh")
					local cmd
					if isWindows then
						cmd = '"'..quoteParam(exeFilename)..' '..quoteParam(tmpFileName)..'"'
					else
						cmd = 'sh '..quoteParam(exeFilename)..' '..quoteParam(tmpFileName)
					end
					for fileName in eatSetContainer(dirsTable):_pairs() do
						tmpFile:write(quoteParam(fileName), "\n")
					end
					tmpFile:close()
					local p = io.popen(cmd)
					if p then
						for line in p:lines() do
							logFn("mapFiles2: line = ", line)
							local fileName = string.gsub(line, '\\', '/')
							logFn("mapFiles2: fileName = ", fileName)
							if ((fileName:len() > 0) and doFilter(fileName)) then
								logFn("mapFiles2: fileName (nach Filter) = ", fileName)
								result:insert(mapFn(fileName))
							end
						end
						p:close()
					end
				end
		end
		
		------------------------------------------------------------------------------------
		------------------------------------------------------------------------------------
		if (type(dirs) == "string") then
			doIt({dirs})
		elseif (type(dirs) == "table") then
			doIt(dirs)
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	mapFiles
	--	
	------------------------------------------------------------------------------------
	mapFiles = function(dir, mapFn, filterFn, subDirs)
		if (not isWindows) then
			--[[
			local doFilter = global.misc.buildFilterFn(filterFn)
			local result = eatSetEx.new()
			local tmpFileName = functionClass.getTempFilename()
			local cmd = (subDirs == true) and string.format('ls "%s" -A -R -U -1 -F -G > "%s"', dir, tmpFileName)
				or string.format('ls "%s" -A -U -1 -F -G > "%s"', dir, tmpFileName)
			local f = io.popen(cmd)
			if f then
				f:close()
				local file = functionClass.openFile(tmpFileName)
				if file then
					local dir = ""
					for line in file:lines() do
						if (line ~= "") then
							if string.ends(line, ":") then
								dir = line:gsub(":", "").."/"
							elseif not string.ends(line, "/") then
								local fileName = string.gsub(dir..line, '\\', '/')
								if ((fileName:len() > 0) and doFilter(fileName)) then
									result:insert(mapFn(fileName))
								end
							end
						end
					end
					file:close()
					functionClass.deleteFile(tmpFileName)
				end
			end
			return result
			]]
			return functionClass.mapFiles2(dir, mapFn, filterFn)
		else
			local doFilter = global.misc.buildFilterFn(filterFn)
			local result = eatSetEx.new()
			local cmd = (subDirs == true) and string.format([[dir "%s" /b /s /a-d]], dir) or string.format([[dir "%s" /b /a-d]], dir)
			logFn("fileutils.mapFiles - cmd = ", cmd)
			local f = io.popen(cmd)
			if f then
				for line in f:lines() do
					local fileName = string.gsub(line, '\\', '/')
					if ((fileName:len() > 0) and doFilter(fileName)) then
						result:insert(mapFn(fileName))
					end
				end
				f:close()
			end
			
			return result
		end
	end,

	------------------------------------------------------------------------------------
	--	getFiles
	--	
	------------------------------------------------------------------------------------
	getFiles = function(dir, filterFn, includeDirName)
		if global.misc.getValue(includeDirName, true) then
			return functionClass.mapFiles(dir, function(fileName) return string.format([[%s%s]], dir, fileName) end, filterFn)
		else
			return functionClass.mapFiles(dir, function(fileName) return fileName end, filterFn)
		end
	end,

	------------------------------------------------------------------------------------
	--	listFiles
	--	
	------------------------------------------------------------------------------------
	listFiles = function(baseDir, includeSubDirs, whiteList, blackList)
		includeSubDirs = global.misc.getValue(includeSubDirs, true)
		whiteList = eatSetContainer(whiteList)
		blackList = eatSetContainer(blackList)
		
		local function filterFn(fileName)
			local result = false
			if (#whiteList > 0) then
				for pattern in whiteList:_pairs() do
					if (string.ends(fileName, pattern)) then
						result = true
						break
					end
				end
			else
				result = true
			end
			if (result and (#blackList > 0)) then
				for pattern in blackList:_pairs() do
					if (string.ends(fileName, pattern)) then
						result = false
						break
					end
				end
			end
			
			return result
		end
		---------------------------------------------------
		local lenBaseDir = string.len(baseDir)
		local function doIt(subDir)
			local result = eatSetEx.new()
			if includeSubDirs then
				local subDirs = functionClass.lisDirs(subDir)
				
				for dir in subDirs:_pairs() do
					local models = doIt(dir)
					for model in models:_pairs() do
						result:insert(model)
					end
				end
			end
			
			local files = functionClass.getFiles(subDir, filterFn)
			for file in files:_pairs() do
				result:insert(string.sub(file, lenBaseDir + 1, string.len(file)))
			end
			
			return result
		end
		
		return doIt(baseDir)
	end,
	-------------------------------------------------------------------------------------------------------
	--	openFile
	--	
	-------------------------------------------------------------------------------------------------------
	openFile = function(fileName, mode, deleteOnClose)
		-------------------------------------------------------------------------------------------------------
		--	mode:
		--	"r"		-	Read-only mode and is the default mode where an existing file is opened.
		--	"w"		-	Write enabled mode that overwrites the existing file or creates a new file.
		--	"a"		-	Append mode that opens an existing file or creates a new file for appending.
		--	"r+"	-	Read and write mode for an existing file.
		--	"w+"	-	All existing data is removed if file exists or new file is created with read write permissions.
		--	"a+"	-	Append mode with read mode enabled that opens an existing file or creates a new file.
		-------------------------------------------------------------------------------------------------------
		local file = io.open(fileName, mode or "r")
		if file then
			return setmetatable({}, {
				__index = {
					-------------------------------------------------------------------------------------------------------
					--	close
					-------------------------------------------------------------------------------------------------------
					close = function(self)
						file:close()
						if (deleteOnClose == true) then
							functionClass.deleteFile(fileName)
						end
					end,
					-------------------------------------------------------------------------------------------------------
					--	flush
					--	Clears the default output buffer.
					-------------------------------------------------------------------------------------------------------
					flush = function(self)
						file:flush()
					end,
					-------------------------------------------------------------------------------------------------------
					--	lines
					--	Provides a generic for loop iterator that loops through the file and closes the file in the end,
					--	in case the file name is provided or the default file is used and not closed in the end of the loop.
					-------------------------------------------------------------------------------------------------------
					lines = function(self)
						return file:lines()
					end,
					-------------------------------------------------------------------------------------------------------
					--	read
					-------------------------------------------------------------------------------------------------------
					read = function(self, arg)
						return file:read(arg)
					end,
					-------------------------------------------------------------------------------------------------------
					--	seek(whence, offset)
					--	Whence parameter is "set", "cur" or "end". Sets the new file pointer with the updated file
					--	position from the beginning of the file. The offsets are zero-based in this function. The offset
					--	is measured from the beginning of the file if the first argument is "set"; from the current
					--	position in the file if it's "cur"; or from the end of the file if it's "end". The default
					--	argument values are "cur" and 0, so the current file position can be obtained by calling this
					--	function without arguments.				
					-------------------------------------------------------------------------------------------------------
					seek = function(self, whence, offset)
						file:seek(whence, offset)
					end,
					-------------------------------------------------------------------------------------------------------
					--	write
					-------------------------------------------------------------------------------------------------------
					write = function(self, ...)
						local args = {...}
						return file:write(table.unpack(args))
					end,
				},
				__newindex = function(self, key, value) end,
				__metatable = false,
			})
		end
	end,
	-------------------------------------------------------------------------------------------------------
	--	getTempFilename
	--	
	-------------------------------------------------------------------------------------------------------
	getTempFilename = function()
		local fileName = _pcall(os.tmpname)
		if fileName then
			logFn("fileName = ", fileName)
			return fileName
		else
			local tmpIndex = 128
			fileName = string.format("%s%d.tmp", global.modPath, tmpIndex)
			while functionClass.fileExists(fileName) do
				tmpIndex = tmpIndex + 1
				fileName = string.format("%s%d.tmp", global.modPath, tmpIndex)
			end
			return fileName
		end
	end,
	-------------------------------------------------------------------------------------------------------
	--	openTempFile
	--	Erstellt und öffnet eine temporäre Datei.
	-------------------------------------------------------------------------------------------------------
	openTempFile = function(deleteOnClose)
		local fileName = functionClass.getTempFilename()
		return functionClass.openFile(fileName, "w+", deleteOnClose), fileName
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
