--- Laneutil
-- @author Enno Sylvester
-- @copyright 2017, 2019, 2023
-- @module eatglobal.laneutil

local laneutil = require "laneutil"

local historie = [[
	Version 1.7:
	 - Changed: Einführung Historie
	 - Changed: 
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 7,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local arr3
local math
local vec3
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		arr3 = global.arr3
		math = global.math
		vec3 = global.vec3
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	calcScale
------------------------------------------------------------------------------------
local function calcScale(dist, angle)
	if (angle < .001) then
		return dist
	end
	
	local pi = 3.14159
	local pi2 = pi / 2	
	local sqrt2 = 1.41421
	
	local scale = 1.0
	if (angle >= pi2) then 
		scale = 1.0 + (sqrt2 - 1.0) * ((angle - pi2) / pi2)
	end

	return .5 * dist / math.cos(.5 * pi - .5 * angle) * angle * scale
end

------------------------------------------------------------------------------------
--	calcEdge
------------------------------------------------------------------------------------
local function calcEdge(points)
	local p0 = arr3.new(points[1])
	local p1 = arr3.new(points[2])
	local t = p1 - p0
	return eatSetEx.new({p0, p1, t, t})
end

------------------------------------------------------------------------------------
--	calcLane
------------------------------------------------------------------------------------
local function calcLane(input)
	local p0 = arr3.new(input[1])
	local p1 = arr3.new(input[2])
	
	local t0 = arr3.new(input[3]):normalize()
	local t1 = arr3.new(input[4]):normalize()
	
	local scale = calcScale(p0:distance(p1), t0:angleUnit(t1))
	
	t0 = t0 * scale
	t1 = t1 * scale
	
	return {p0, t0, p1, t1}
end

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	vecToPoint
	------------------------------------------------------------------------------------
	vecToPoint = function(v)
		return arr3.new(v)
	end,
	------------------------------------------------------------------------------------
	--	pointToVec
	------------------------------------------------------------------------------------
	pointToVec = function(pt)
		return vec3.new(pt)
	end,
	------------------------------------------------------------------------------------
	--	makeLanes
	------------------------------------------------------------------------------------
	makeLanes = function(input)
		local result = eatSetEx.new()
		
		for i = 1, #input do
			local lane = calcLane(input[i])
			local width = input[i][5]
			result:inert({lane[1], lane[2], width})
			result:inert({lane[3], lane[4], width})
		end	
		return result
	end,
	------------------------------------------------------------------------------------
	--	makeLane
	------------------------------------------------------------------------------------
	makeLane = function(source, dest, laneWidth)
		--[[
		local t = arr3.sub(dest, source)
		return { source, dest, t, t, laneWidth }
		]]
		result = calcEdge({source, dest})
		result:insert(laneWidth)
		return result
	end,
	------------------------------------------------------------------------------------
	--	makeLaneEx
	------------------------------------------------------------------------------------
	makeLaneEx = function(points, laneWidth)
		if (type(points) == 'table') and (#points == 4) then
			local t1 = arr3.sub(points[2], points[1])
			local t2 = arr3.sub(points[4], points[3])
			return { points[2], points[3], t1, t2, laneWidth}
		else
			return {}
		end
	end,
	------------------------------------------------------------------------------------
	--	makeEdges
	------------------------------------------------------------------------------------
	makeEdges = function(input)
		local result = eatSetEx.new()
		for i = 1, #input do
			local lane = calcLane(input[i])
			result:inert({lane[1], lane[2]})
			result:inert({lane[3], lane[4]})
		end	
		return result
	end,
	------------------------------------------------------------------------------------
	--	edgeFromPoints
	--	Erwartet eine Tabelle mit 2 Punkten: edgeFromPoints({x1, y1, z1}, {x2, y2, z2})
	--	Rückgabe eatSetEx
	------------------------------------------------------------------------------------
	edgeFromPoints = function(points)
		return functionClass.makeEdges({ calcEdge(points) })
	end,
	------------------------------------------------------------------------------------
	--	edgesFromPoints
	--	Erwartet eine Tabelle mit Einträgen a 2 Punkten:
	--	edgesFromPoints({
	--			{{x1, y1, z1}, {x2, y2, z2}},
	--			{{x3, y3, z3}, {x4, y4, z4}},
	--			{{x5, y5, z5}, {x6, y6, z6}},
	--			...
	--		})
	--	Rückgabe eatSetEx
	------------------------------------------------------------------------------------
	edgesFromPoints = function(pointsList)
		local edges = eatSetEx.new()
		for points in eatSetContainer(pointsList):_pairs() do
			edges:insert(calcEdge(points))
		end
		return functionClass.makeEdges(edges)
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB, laneutil}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

