--- maputil
-- @author Enno Sylvester
-- @copyright 2018, 2019, 2021
-- @module eatglobal.maputil

local historie = [[
	Version 1.0:
	 - Release
	Version 1.1:
	 - Added: Added Fields "__debug__" and "historie" to result-table maputil.
	 - Changed: Adjustments to modified arr3 library.
	 - Changed: Functions with table returns aka lists, returns now as eatSetEx.
	Version 1.2:
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 2,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local arr3
local math
local transf

------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		arr3 = global.arr3
		transf = global.transf
		math = global.math
	end,
}

------------------------------------------------------------------------------------
--	
------------------------------------------------------------------------------------
function libResult.createMapping(radius)
	local mapping = {
			radius = radius,
		}
	
	--[[
		################################################################
	]]
	mapping.pointData = function(p)
		local function calcData(pt)
			local x, y, z = pt:toParams()
			if (radius == 0) then
				return {
					origPoint = pt:copy(),
					mappedPoint = pt:copy(),
					radius = 0,
					radiusAtXPosition = 0,
					aRad = 0,
					yLength = y,
					heigth = z,
					tangent = arr3.new(0, 1, 0),
					scaleFactor = 1,
					meshScale = arr3.new(1, 1, 1),
				}
			else
				local r = radius + x
				local aRad = y / radius
				local scaleFactor = r / radius
				local yLength = aRad * r
				-- Daten
				return {
					origPoint = pt:copy(),
					mappedPoint = arr3.new(math.cos(aRad) * r - radius, math.sin(aRad) * r, z),
					radius = radius,
					radiusAtXPosition = r,
					aRad = aRad,
					yLength = yLength,
					heigth = z,
					tangent = arr3.new(-math.sin(aRad), math.cos(aRad), 0),
					scaleFactor = scaleFactor,
					meshScale = arr3.new(1, scaleFactor, 1),
				}
			end
		end
		local result = {
			data = calcData(arr3.new(p))
		}

		-- Hilfsfunktionen
		local function angleArg(arg)
			return math.acos(math.clamp(arg, -1, 1))
		end

		local function checkBoolArray(v, minCount)
			local result = eatSetEx.new()
			minCount = (type(minCount) == "number") and math.abs(math.round(minCount)) or 3
			if (type(v) == "table") then
				for i = 1, minCount do
					result:insert((type(v[i]) ~= "boolean") and false or v[i])
				end
			else
				result:insert(false, minCount)
			end
			return result
		end
	
		-- Funktionen
		result.applyRotation = function (rotation)
				return arr3.new(rotation[1], rotation[2], rotation[3] + result.data.aRad)
			end

		result.getLengthTo = function (p1)
				local function calcTotalLength(yLength, z)
					local fac = (yLength < 0) and -1 or 1
					return math.sqrt((yLength * yLength) + (z * z)) * fac
				end
				local p1Data = calcData(arr3.new(p1))
				return calcTotalLength(p1Data.yLength - result.data.yLength, p1Data.heigth - result.data.heigth)
			end
			
		result.getYLengthTo = function (p1)
				local p1Data = calcData(arr3.new(p1))
				return p1Data.yLength - result.data.yLength
			end

		result.calcZData = function(p1, slopes)
				local p1Data = calcData(arr3.new(p1))
				local _slopes = checkBoolArray(slopes)
				local height = p1Data.heigth - result.data.heigth
				local angle = angleArg(height)
				
				local slopeCorrection = 0.625
				local fac = (_slopes[1] == true) and math.addPercent(1, slopeCorrection) or 1
				fac = (_slopes[2] == true) and math.addPercent(fac, slopeCorrection) or fac
				local length = result.getLengthTo(p1) * fac
				
				return {
					angle = angle,
					length = length,
					height = height,
					argAndAngel = angle * length,
					slopes = _slopes,
				}
			end
			
		result.calcDirectionVectors = function (p1, slopes)
				local t0, t1
				local zData = result.calcZData(p1, slopes)
				if (radius == 0) then
					t0 = arr3.new(p1) - result.data.origPoint
					t1 = arr3.new(t0)
				else
					local p1Data = calcData(arr3.new(p1))
					t0 = result.data.tangent * zData.length
					t1 = p1Data.tangent * zData.length
				end
				if (zData.height ~= 0) then
					t0[3] = (zData.slopes[1] == true) and zData.argAndAngel or zData.height
					t1[3] = (zData.slopes[2] == true) and zData.argAndAngel or zData.height
				end
				return t0, t1
			end

		result.calcCompleteMapping = function(p1, slopes)
				local p1Data = calcData(arr3.new(p1))
				local t0, t1 = result.calcDirectionVectors(p1, slopes)
				return result.data.mappedPoint, t0, p1Data.mappedPoint, t1
			end
			
		result.getMeshTransf = function (doYScale, additionalScale, additionalRotate, forceXScale)
				local _doYScale = (doYScale ~= nil) and doYScale or true
				local scale = (_doYScale == true) and result.data.meshScale or arr3.new(1)
				if (forceXScale == true) then
					scale = arr3.new(scale[2], scale[1], scale[3])
				end
				_additionalRotate = (type(additionalRotate) == 'table') and additionalRotate or {0, 0, 0}
				local _result = transf.scaleRotXYZTransl(
					scale:toVec3(),
					result.applyRotation(_additionalRotate):toVec3(),
					result.data.mappedPoint:toVec3()
				)
				if (type(additionalScale) == 'table') then
					_result = transf.setScale(_result, additionalScale)
				end
				return _result
			end
			
		return result
	end
	
	mapping.mapPoint = function (p)
			local pointData = mapping.pointData(p)
			return pointData.data.mappedPoint
		end
	
	mapping.mapPoints = function (points)
			local result = eatSetEx.new()
			for i = 1, #points do
				result:insert(mapping.mapPoint(points[i]))
			end
			return result
		end
		
	return mapping
end

return libResult