--- utils
-- @author Enno Sylvester
-- @copyright 2017, 2018, 2019, 2020, 2023
-- @module eatglobal.utils

local historie = [[
	Version 1.14:
	 - Added:	getFileExtension(fileName, includeDot)
	 - Added:	getFileNameAndExtension(fileName, includePath)
	 - Added:	getFileFolders(fileName)
	Version 1.15:
	 - Added:	splitString(s, pattern)
	 - Added: compareVersion(version1, version2)
	 - Added: isValidVersion(version)
	 - Added: mapFiles(dir, fnc, filterFn)
	Version 1.16:
	 - Changed:	compareVersion greift auf globale compareVersion aus eatglobal_intern zu
	Version 1.17:
	 - Fixed:	interne Überarbeitungen
	 - Changed:	compareVersion greift auf globale eatCompareVersion aus eatglobal_vx_yy zu
	Version 1.18:
	 - Fixed: function _readFileToEnv -> "file, err = loadfile(fileName, "tb", env)" to "local file, err = loadfile(fileName, "tb", env)"
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 18,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local serialize
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		serialize = global.serialize
	end,
}

------------------------------------------------------------------------------------
--	
------------------------------------------------------------------------------------
--[[
			##########################################
			environment-tools
			##########################################
]]

local function init()
	if (type(libResult.env) ~= "table") then
		libResult.env = {}
	end
	libResult.env.isWinOS = (os.getenv("OS") == "Windows_NT")
end

init ()

local function concatTable( tbl, CR )
	local doCR = CR or false
	local stack = {}
	for i = 1, #tbl do
		stack[#stack + 1] = libResult.toString(tbl[i])
		if (doCR) and (i < #tbl) then
			stack[#stack + 1] = "\n"
		end
	end
	return table.concat(stack)
end

--- log mit Präfix
local createLogFn
local createLogFnCounter
function libResult.createLogFn(prefix, traceback)
	if (createLogFn == nil) then
		createLogFn = getVar("eat.intern.createLogFn")
	end
	return createLogFn(prefix, traceback)
end

function libResult.createLogFnCounter(prefix, traceback)
	if (createLogFnCounter == nil) then
		createLogFn = getVar("eat.intern.createLogFnCounter")
	end
	return createLogFnCounter(prefix, traceback)
end

function libResult.getModulFileName ( level )
	if (level == nil) then
		level = 1
	end
	local info = debug.getinfo(level, "S")
	if (info ~= nil) then
		return string.gsub(string.gsub(info.source, '\\', '/'), '@', "")
	else
		return ""
	end
end

function libResult.getModPath()
	local function debugPath ()
		local level = 1
		local info
		
		repeat
			info = debug.getinfo(level, "S")
			level = level + 1
		until (info == nil)
		if (level > 1) then
			return string.gsub(string.gsub(debug.getinfo(level - 2, 'S').source, '\\', '/'), '@', "")
		else
			return ""
		end
	end

	local result = debugPath ()
	local i = result:find("/res/")
	if i then
		return result:sub(1, i)
	else
		return result:match("(.*[/\\])")
	end
end

------------------------------------------------------------------------------------
--	Extrahiert den Dateinamen (incl. Extension) aus 'fileName'.
--
--	"c:/werner/werner.txt"	-->	"werner.txt"
--	"c:/werner/.txt"				-->	".txt"
--	"c:/werner/."						-->	"."
--	"c:/werner/werner"			-->	"werner"
--	"c:/werner/"						-->	""
--	"c:/"										-->	""
------------------------------------------------------------------------------------
function libResult.getFileName(fileName)
	if eatIsValidId(fileName) then
		local sLen = fileName:len()
		for i = sLen, 1, -1 do
			if (i == 1) then
				return fileName
			else
				local char = fileName:sub(i, i)
				if ((char == "/") or (char == ":")) then
					return fileName:sub(i + 1, sLen)
				end
			end
		end
	end
	return ""
end

------------------------------------------------------------------------------------
--	Ermittelt Pfad, Dateiname  und Dateiendung aus 'fileName'.
--	Rückgabe ist eine Tabelle mit den Einträgen 'path', 'name' und 'ext'.
------------------------------------------------------------------------------------
function libResult.getFileData(fileName)
  local fName = libResult.getFileName(fileName)
  local data = string.eat.split(fName, ".")
  return {
    name = data[1] or "",
    ext = data[2] or "",
    path = (fName ~= "") and fileName:sub(1, fileName:find(fName, 1, true) - 1) or fileName,
  }
end

------------------------------------------------------------------------------------
--	Extrahiert die Dateiendung aus 'fileName'.
--	Ist 'includeDot' == true, so wird der '.' mit ausgegeben (z.B. ".lua", ".txt").
------------------------------------------------------------------------------------
function libResult.getFileExtension(fileName, includeDot)
	local data = libResult.getFileData(fileName)
	return ((includeDot == true) and (data.ext ~= "")) and "."..data.ext or data.ext
end

------------------------------------------------------------------------------------
--	Gibt den Dateinamen ud die Dateierweiterung zurück.
--	Ist 'includePath' == true wird der Dateiname inclusive Pfad zurückgegeben.
--
--	fileName = "mods/eat1963/jaaaaa/res/scripts/www.lua"; includePath = false oder nil
--	--> www			lua
--	fileName = "mods/eat1963/jaaaaa/res/scripts/www.lua"; includePath = true
--	--> mods/eat1963/jaaaaa/res/scripts/www			lua
------------------------------------------------------------------------------------
function libResult.getFileNameAndExtension(fileName, includePath)
	local data = libResult.getFileData(fileName)
  return (includePath ~= true) and data.name or data.path..data.name, data.ext
end

function libResult.fileNameIncludeSuffix(fileName, suffix, includePath)
	local fName, fExt = libResult.getFileNameAndExtension(fileName, includePath)
	return fName..suffix.."."..fExt
end

function libResult.getFileFolders(fileName)
	local result = {}
	local currentIndex = 1
	local currentFound = fileName:find("/", currentIndex, true)
	while currentFound do
		if (currentFound > currentIndex) then
			result[#result + 1] = fileName:sub(currentIndex, currentFound - 1)
		end
		currentIndex = currentFound + 1
		currentFound = fileName:find("/", currentIndex, true)
	end
	return result
end

--[[
			##########################################
			table-tools
			##########################################
]]

--- Kopiert eine Tabelle incl. der metatable
-- @param tbl
-- @return Kopie von tbl
-- Achtung: In Kopien von Proxytabellen kann nicht geschrieben oder gelöscht werden.
function libResult.copyTable (tbl)
	local function doCopy (t)
		local result
		if (type(t) == 'table') then
			result = {}
			for key, value in next, t, nil do
				result[doCopy(key)] = doCopy(value)
			end
			setmetatable(result, doCopy(getmetatable(t)))
		else -- number, string, boolean, etc
			result = t
		end
		return result
	end
	return doCopy(tbl)
end

--- Create a ordered table
-- @return {table} a table with ordered key value iteration
-- @author Oskar Eisemuth
-- @copyright 2017
 function libResult.createOrderedTable()
	local mt = {
		__index = {
			_keyorder = {},
			_delete = function(self, key)
				if self[key] == nil then
					return
				end
				self[key] = nil
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						table.remove(keyorder, i)
					end
				end
			end
		},
		__newindex = function(self, key, value)
			rawset(self, key, value)
			local keyorder = self._keyorder
			keyorder[#keyorder+1] = key
		end,
		__pairs = function(tbl)
			local i = 0
			local function pairs_iter(self)
				i = i + 1
				local key = self._keyorder[i]
				if (key) then
					return key, self[key]
				end
				return
			end
			return pairs_iter, tbl, nil
		end
	}
  return setmetatable({}, mt)
end

--- Erzeugt eine Proxy-Tabelle.
--	Alle Zugriffe werden direkt auf den 'master' umgeleitet
-- @libResult.proxyTable
-- @param master
-- @return proxy
function libResult.proxyTable(master)
  local mt = {
    __index = master,
    __newindex = master,
		__len = function(self)
			return #master
		end,
		__pairs = function(self)
      return function(self, key)
        local value
        key, value = next(master, key)
        if (value ~= nil) then
          return key, value
        end
      end, self, nil
    end,
		__ipairs = function(self)
			return function(self, i)
				i = i + 1
				local value = master[i]
				if (value ~= nil) then
					return i, value
				end
			end, self, 0
    end,
  }
  return setmetatable({}, mt)
end

local function buildFilterFn(filterFn)
	if (type(filterFn) == "function") then
		return filterFn
	elseif (type(filterFn) == "table") then
		return function(t, key, value)
			for _, fn in ipairs(filterFn) do
				if not fn(t, key, value) then
					return false
				end
			end
			return true
		end
	else
		return function(t, key, value)
			return true
		end
	end
end
--- Äquivalent zu pairs (ab Version 1.10)
--	Iteriert über eine Tabelle, wobei nur Einträge berücksichtigt werden, welche
--		filterFn mit "true" passiert haben
-- Nutzung:
-- for key, value in filterPairs (tbl, filterFn) do
-- end
--	Beispiel filterFn:
--	local function filterFn(tbl, key, value)
--		return (value ~= "1000") and value ~= "760"
--	end
function libResult.filterPairs ( tbl, filterFn )
	-- collect the keys
	local keys = {}
	for k in pairs(tbl) do
		keys[#keys + 1] = k
	end

	local doFilter = buildFilterFn(filterFn)
	-- return the iterator function
	local i = 0
	return function()
		i = i + 1
		while keys[i] and (not doFilter(tbl, keys[i], tbl[keys[i]])) do
			i = i + 1
		end
		
		if keys[i] then
			return keys[i], tbl[keys[i]]
		end
	end
end

--- Äquivalent zu pairs
-- Iteriert über eine Tabelle, wobei die Keys in sortierter Reihenfolge aufgerufen werden.
-- Nutzung:
-- for key, value in sortPairs (tbl) do
-- end
-- ab Version 1.10 mit filterFn
function libResult.sortPairs ( tbl, orderFn, filterFn )
	local function noFilter(tbl, key, value)
		return true
	end
	-- collect the keys
	local keys = {}
	for k in pairs(tbl) do
		keys[#keys + 1] = k
	end

	-- if order function given, sort by it by passing the table and keys a, b,
	-- otherwise just sort the keys 
	if orderFn then
		table.sort(keys, function(a,b) return orderFn(tbl, a, b) end)
	else
		table.sort(keys)
	end

	local doFilter = buildFilterFn(filterFn)
	-- return the iterator function
	local i = 0
	return function()
		i = i + 1
		while keys[i] and (not doFilter(tbl, keys[i], tbl[keys[i]])) do
			i = i + 1
		end
		
		if keys[i] then
			return keys[i], tbl[keys[i]]
		end
	end
end

function libResult.ipairs(tbl)
	-- return the iterator function
	local i = 0
	return function()
		i = i + 1
		if tbl[i] then
			return tbl[i]
		end
	end
end
--- Gibt eine sortierte Kopie von tbl zurück
--	
-- @libResult.sortTable
-- @param tbl
-- @param orderFn
-- @return die sortierte Tabelle, oder das Original, wenn orderFn == nil
function libResult.sortTable ( tbl, orderFn )
	if (type(orderFn) == "function") then
		local temp = {}
		for k in libResult.sortPairs(tbl, orderFn) do
			temp[k] = tbl[k]
		end
		return temp
	else
		return tbl
	end
end

--- Ergänzt Tabelle "tbl" mit Werten aus Tabelle "source"
--
-- @libResult.fillTable
-- @param tbl
-- @param source
-- @return die ergänzte Tabelle, oder das Original, wenn source keine Tabelle
function libResult.fillTable (tbl, source)
	local function merge(a, b)
		if (type(a) == 'table') and (type(b) == 'table') then
			for k,v in pairs(b) do
				if (type(v) == 'table') and (type(a[k] or false) == 'table') then
					merge(a[k],v)
				else
					a[k] = v
				end
			end
		end
		return a
	end
	
	if (type(source) == 'table') then
		return merge(tbl, source)
	else
		return tbl
	end
end

function libResult.indexOf(tbl, element)
	if (type(tbl) == 'table') and (type(element) ~= nil) then
		for i = 1, #tbl do
			if (tbl[i] == element) then
				return i
			end
		end
	end
	return 0
end

function libResult.isEmptyTable(tbl)
	for k in pairs(tbl) do
		return false
	end
	
	return true
end
--[[
			##########################################
			Filetools
			##########################################
]]

--- prüft, ob eine Datei vorhanden ist
-- @libResult.fileExists
-- @param fileName
-- @return true, wenn vorhanden
function libResult.fileExists (fileName)
	local file = io.open(fileName, "r")
	if file then
		file:close()
		return true
	end
	return false
end

--- Daten aus Datei einlesen
-- @libResult.readData
-- @param fileName
-- @return Tabelle mit eingelesene Datan oder eine leere Tabelle
local function _readFileToEnv( fileName, env)
	if (env ~= nil) then
		setmetatable(env, {__index = _G})
		local file, err = loadfile(fileName, "tb", env)
		if (file~= nil) then
			file()
			return true
		else
			libResult.logFn("libResult.readData - File '", fileName, "' Error: ", err)
			return false, err
		end
	end
	
	return false, ""
end

--- Daten aus Datei einlesen
-- @libResult.readData
-- @param fileName
-- @param dataFormat: "TEXT", "BLANK", "FUNC_DATA", "RETURN"
function libResult.readData (fileName, dataFormat)
	dataFormat = dataFormat or "STRUCT"
	
	if (dataFormat == "FUNC_DATA") then
		local env = {
			data = function()
				return {}
			end
		}
		local result, err = _readFileToEnv(fileName, env)
		if result then
			if (env.data ~= nil) then
				return env.data()
			else
				libResult.logFn("libResult.readData - File '", fileName, "' contains no function 'data()'.")
				return {}
			end
		else
			return {}, err
		end
	elseif (dataFormat == "TEXT") then
		local file = io.open(fileName, "r")
		if file then
			local s = file:read("*a")
			file:close()
			return s
		else
			return nil, "Unable to open ["..fileName.."]."
		end
	else
		local file, err = loadfile(fileName)
		if (file ~= nil) then
			local result = file()
			if (type(result) ~= "table") then
				result = {}
				libResult.logFn("libResult.readData - File '", fileName, "' is not a valid datafile. Maybe he's corrupt?!")
			end
			return result
		else
			if libResult.fileExists(fileName) then
				libResult.logFn("libResult.readData - File '", fileName, "' Error: ", err)
			end
			return {}, err
		end
	end
end

--- Liest Datenstruktur aus fileName.
--	Ist die Datei vorhanden, aber enthält keine Datenstruktur,
--	so wird die Datei geleert und mit einer leeren Datenstruktur gefüllt.
-- @libResult.readDataEx
-- @param fileName
-- @return Tabelle mit eingelesene Datan oder eine leere Tabelle
function libResult.readDataEx(fileName)
	local file, err = loadfile(fileName)
	if (file ~= nil) then
		local result = file()
		if (type(result) ~= "table") then
			libResult.saveData({}, fileName)
			libResult.logFn("libResult.readDataEx - File '", fileName, "' is not a valid datafile. Maybe he's corrupt?!")
			return {}
		else
			return result
		end
	else
		if libResult.fileExists(fileName) then
			libResult.logFn("libResult.readDataEx - File '", fileName, "' Error: ", err)
			libResult.saveData({}, fileName)
		end
		return {}, err
	end
end

--- Daten in Datei schreiben
-- @libResult.saveData
-- @param data
-- @param fileName
-- @param dataFormat: "TEXT", "BLANK", "FUNC_DATA", "RETURN"
-- @param append - boolean (nur bei dataFormat == "TEXT")
function libResult.saveData (data, fileName, dataFormat, append)
	local file
	local s
	local dataFormat = dataFormat or "RETURN"
	if (dataFormat == "TEXT") then
		local fileMode = (append == true) and "a" or "w"
		file = io.open(fileName, fileMode)
	else
		file = io.open(fileName, "w")
	end

	if file then
		if (dataFormat == "BLANC") then
			s = serialize.serializeBlanc(data)
		elseif (dataFormat == "FUNC_DATA") then
			s = serialize.serializeDatafmt(data)
		elseif (dataFormat == "TEXT") then
			s = data
		else
			s = serialize.serializeReturnfmt(data)
		end
		
		file:write(s)
		file:close()
	end
end

function libResult.deleteFile(fileName)
	local function quote( var )
		return '"'..string.gsub(var, '/', '\\')..'"'
	end
		
	if libResult.fileExists (fileName) then
		local cmd = nil
		if libResult.env.isWinOS then
			cmd = '"del '..quote(fileName)..'" '
		else
			cmd = '"rm '..quote("./"..fileName)..'" '
		end
		p = io.popen(cmd)
		if p then
			p:close()
		end
	end
end

function libResult.getSubDirs(dir)
	local result = eatSetEx.new()
	local f = io.popen(string.format([[dir "%s" /b /ad]], dir))
	if f then
		for s in f:lines() do
			if (s.len() > 0) then
				result:insert(string.format([[%s%s/]], dir, s))
			end
		end
		f:close()
	end
	
	return result
end

function libResult.mapFiles(dir, fnc, filterFn)
	filterFn = filterFn or function(fileName) return true end
	local result = eatSetEx.new()
	local f = io.popen(string.format([[dir "%s" /b /a-d]], dir))
	if f then
		for fileName in f:lines() do
			if ((fileName:len() > 0) and filterFn(fileName))then
				result:insert(fnc(fileName))
			end
		end
		f:close()
	end
	
	return result
end

function libResult.getFiles(dir, filterFn)
	local function fnc(fileName)
		return string.format([[%s%s]], dir, fileName)
	end
	return libResult.mapFiles(dir, fnc, filterFn)
end

function libResult.listFiles(baseDir, includeSubDirs, whiteList, blackList)
	includeSubDirs = (type(includeSubDirs) == "boolean") and includeSubDirs or true
	whiteList = eatSetEx.new(whiteList or {})
	blackList = eatSetEx.new(blackList or {})
	local function filterFn(fileName)
		local result = false
		if (#whiteList > 0) then
			for pattern in whiteList:_pairs() do
				if (string.ends(fileName, pattern)) then
					result = true
					break
				end
			end
		else
			result = true
		end
		if (result and (#blackList > 0)) then
			for pattern in blackList:_pairs() do
				if (string.ends(fileName, pattern)) then
					result = false
					break
				end
			end
		end
		
		return result
	end
	local lenBaseDir = string.len(baseDir)
	local function doIt(subDir)
		local result = eatSetEx.new()
		if includeSubDirs then
			local subDirs = libResult.getSubDirs(subDir)
			
			for dir in subDirs:_pairs() do
				local models = doIt(dir)
				for model in models:_pairs() do
					result:insert(model)
				end
			end
		end
		
		local files = libResult.getFiles(subDir, filterFn)
		for file in files:_pairs() do
			result:insert(string.sub(file, lenBaseDir + 1, string.len(file)))
		end
		
		return result
	end
	
	return doIt(baseDir)
end

--[[
			##########################################
			math-tools
			##########################################
]]

function libResult.getArrayIndex(array, params, compareFn)
	for index, value in ipairs(array) do
		if compareFn(value, params) then
			return index
		end
	end
	return 0
end

function libResult.getIndex(tbl, params, compareFn)
	return libResult.getArrayIndex(tbl, params, compareFn)
end

function libResult.toString (var)
	if type(var) == "nil" then
		return "nil"
	elseif type(var) == "boolean" then
		return tostring(var)
	elseif type(var) == "number" then
		return tostring(var)
	elseif type(var) == "string" then
		return var
	elseif type(var) == "table" then
		return serialize.serializeBlanc(var)
	else
		return ""
	end
end

function libResult.round(var)
	return global.math.round(var)
end

--[[
			##########################################
			string-tools
			##########################################
]]

function libResult.isValidString(s)
	return global.misc.isValidId(s)
end

function libResult.splitString(s, pattern)
	return string.eat.split(s, pattern)
end
--[[
			##########################################
			misc-tools
			##########################################
]]

function libResult.concatParams (...)
	return concatTable( {...}, false)
end

function libResult.concatParamsCr (...)
	return concatTable( {...}, true)
end

function libResult.pRequire(modulName)
	return _pRequire(modulName)
end

function libResult.getOrder(userId, modId, modOrder)
	return (userId * 10000) + (modId * 100) + modOrder
end

function libResult.getValue(value, default)
	return global.misc.getValue(value, default)
end

function libResult.dateTimeString()
	return os.date("%d.%m.%Y %X", os.time())
end

function libResult.isValidVersion(version)
	return global.versiondata.isValidVersion(version)
end

function libResult.compareVersion(version1, version2)
	return global.versiondata.isValidVersion(version1, version2)
end

return libResult