--- vec2
-- @author Enno Sylvester
-- @copyright 2019, 2021, 2023
-- @module eatglobal.vec2

local historie = [[
	Version 1.0:
	 - Release
	Version 1.1:
	 - Added: Modulfähigkeit (Erweiterungsmöglichkeiten durch Module) implementiert.
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local arr2
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		arr2 = global.arr2
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local function getVec2(x, y)
	if (type(x) == "table") then
		if ((type(x[1]) == "number") and (type(x[2]) == "number")) then
			-- x is a arr2 or arr3
			return {x = x[1], y = x[2]}
		elseif ((type(x.x) == "number") and (type(x.y) == "number")) then
			-- x is a vec2, vec3 or vec4
			return {x = x.x, y = x.y}
		end
	elseif (type(x) == "number") then
		-- x is a number; We look to see if y are also a number. Otherwise they get x as value
		return {x = x, y = (type(y) == "number") and y or x}
	end
	-- nothing; return default
	return {x = 0, y = 0}
end
	
local mt
mt = {
  __add = function (self, rhs) -- "add" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x + rhs, y = self.y + rhs}, mt)
		else
			local v = getVec2(rhs)
			return setmetatable({x = self.x + v.x, y = self.y + v.y}, mt)
		end
  end,
	__sub = function (self, rhs) -- "sub" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x - rhs, y = self.y - rhs}, mt)
		else
			local v = getVec2(rhs)
			return setmetatable({x = self.x - v.x, y = self.y - v.y}, mt)
		end
  end,
	__mul = function (self, rhs) -- "mul" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x * rhs, y = self.y * rhs}, mt)
		else
			local v = getVec2(rhs)
			return setmetatable({x = self.x * v.x, y = self.y * v.y}, mt)
		end
	end,
	__div = function (self, rhs) -- "div" event handler
		if (type(rhs) == "number") then
			return setmetatable({x = self.x / rhs, y = self.y / rhs}, mt)
		else
			local v = getVec2(rhs)
			return setmetatable({x = self.x / v.x, y = self.y / v.y}, mt)
		end
	end,
	__index = {
		swap = function(self)
			return setmetatable({self.y, self.x}, mt)
		end,
		toArray = function(self)
			return arr2.new(self)
		end,
		toParams = function(self)
			return self.x, self.y
		end,
		toString = function(self)
			return tostring(self)
		end,
	},
	__tostring = function(self)
		return string.format("{x = %.4f, y = %.4f}", self:toParams())
	end,
}

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	new
	--	
	------------------------------------------------------------------------------------
	new = function(x, y)
		return setmetatable(getVec2(x, y), mt)
	end,
	------------------------------------------------------------------------------------
	--	copy
	--	
	------------------------------------------------------------------------------------
	copy = function(vec)
		return functionClass.new(vec.x, vec.y)
	end,
	------------------------------------------------------------------------------------
	--	mt
	--	
	------------------------------------------------------------------------------------
	mt = function(vec)
		return functionClass.new(vec.x, vec.y)
	end,
	------------------------------------------------------------------------------------
	--	mtVecs
	--	
	------------------------------------------------------------------------------------
	mtVecs = function(vecs)
		local result = {}
		for i = 1, #vecs do
			result[#result + 1] = functionClass.mt(vecs[i])
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	add
	--	
	------------------------------------------------------------------------------------
	add = function(a, b)
		return functionClass.new(a.x + b.x, a.y + b.y)
	end,
	------------------------------------------------------------------------------------
	--	sub
	--	
	------------------------------------------------------------------------------------
	sub = function(a, b)
		return functionClass.new(a.x - b.x, a.y - b.y)
	end,
	--- Ex- Importfunktionen
	------------------------------------------------------------------------------------
	--	fromArray
	--	
	------------------------------------------------------------------------------------
	fromArray = function(a)
		return functionClass.new(a[1], a[2])
	end,
	------------------------------------------------------------------------------------
	--	toArray
	--	
	------------------------------------------------------------------------------------
	toArray = function(vec)
		return arr2.fromVec(vec)
	end,
	------------------------------------------------------------------------------------
	--	toParams
	--	
	------------------------------------------------------------------------------------
	toParams = function(vec)
		return vec.x, vec.y
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

local vec2 = require "vec2"
return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB, vec2}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
