local startInit = os.clock()
--- Main
-- @author Oskar Eisemuth, Enno Sylvester (Anpassung)
-- @copyright 2017, 2018, 2019, 2020, 2021, 2022, 2023
-- @module eatglobal
local __debug__ = false

local historie = [[
	Version 2.30:
	 - Added: checkEnvironment() - Alte eatglobal aus aktuellen API-Ordner löschen. "eatglobal_historie.txt" ertsellen.
	Version 2.31:
	 - Changed: Diverse Bugs in API-Dateien beseitigt
	Version 2.32xxx:
	 - Added: global function eatIsValidId(id)
	 - Added: global function eatLogFn(prefix, traceback). Ist Parameter traceback == true, dann Ausgaben mit debug.traceback.
	 - Added: global function eatLogFnCounter(prefix, traceback). Ist Parameter traceback == true, dann Ausgaben mit debug.traceback.
	 - Added: global function eatHasMetaFunction(t, fn)
	 - Added: global function eatUnpack(t)
	 - Changed: Alle Module in "res/scripts/eatglobal/" haben nun den Vorsatz "eat_" um versehentliches Laden gleichnamiger Skripte zu vermeiden.
	 - Changed: Alle Module in "res/scripts/eatglobal/", welche nicht den Vorsatz "eat_" haben, werden gelöscht.
	Version 2.32012:
	 - Added: Modul meshhandler
	 - Changed: Überarbeitung eatglobal.laneutil
	 - Changed: Entfernung Modul materialshandler, da Aufgabe von meshhandler übernommen wird.
]]

local versionsData = {
	majorVersion = 2,
	minorVersion = 32013,
}

local eatglobal = {
	-------------------------------------------------------------------
	--	Different tables, which are filled with functions and/or
	--	data for global access.
	-------------------------------------------------------------------
	misc = {},
	historie = {},
	forceCleanApi = "dummy",
	loadDataFile = "dummy",
	apiPath = "dummy",
	modPath = "dummy",
	-------------------------------------------------------------------
	--	Dump-Function
	-------------------------------------------------------------------
	dmp = function(...)
		local ok, inspect = pcall(require, "inspect")
		if (ok) then
			print(inspect(...))
		elseif (type(eatToString) == "function") then
			local stack = {}
			for _, value in ipairs({...}) do
				stack[#stack + 1] = eatToString(value)
			end
			print(table.concat(stack))
		end
	end,
}

------------------------------------------------------------------------------------
--	Später benötigte Variablen definieren
------------------------------------------------------------------------------------
local moduleFile = ""
local modulePath = ""
local eatglobal_intern
local setVar
local getVar
local fileutils
local versiontools
local apiSettings = {
	module = {
		librarys = {},
	},
}
local splitString
------------------------------------------------------------------------------------
--	eatglobal.misc.buildFilterFn
--	
------------------------------------------------------------------------------------
function eatglobal.misc.buildFilterFn(filterFn)
	------------------------------------------------------------------------------------
	--	noFilterFn(value)
	--	
	------------------------------------------------------------------------------------
	local function noFilterFn(value)
		return true
	end
	------------------------------------------------------------------------------------
	--	filterTableFn(value)
	--	
	--	Aufbau filterFn als Tabelle:
	--	filterFn = {
	--		function_1,
	--		function_2,
	--		...
	--		function_n,
	--		orFunctions = {
	--			function_1,
	--			function_2,
	--			...
	--			function_n,
	--		},
	--	}
	--	
	--	Alle Rückgaben der Funktionen direkt in Tabelle "filterFn" müssen "true" sein.
	--	Ist die Untertabelle "filterFn.orFunctions" vorhanden, so muss zusätzlich
	--	mindestens eine der Funktionen aus "filterFn.orFunctions" "true" zurückgeben.
	--	Es können sowohl nur Funktionen oder nur "filterFn.orFunctions" angegeben
	--	werden.
	------------------------------------------------------------------------------------
	local function filterTableFn(value)
		for key, fun in pairs(filterFn) do
			if (type(fun) == "function") then
				if (not fun(value)) then
					return false
				end
			end
		end
		if (type(filterFn.orFunctions) == "table") then
			for i, fun in ipairs(filterFn.orFunctions) do
				if (type(fun) == "function") then
					if fun(value) then
						return true
					end
				end
			end
			return false
		end
		return true
	end
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	if (type(filterFn) == "function") then
		return filterFn
	elseif (type(filterFn) == "table") then
		return filterTableFn
	else
		return noFilterFn
	end
end
local buildFilterFn = eatglobal.misc.buildFilterFn

------------------------------------------------------------------------------------
--	eatglobal.misc.isValidId
--	Prüft 'id' auf gültigen Wert.
------------------------------------------------------------------------------------
function eatglobal.misc.isValidId(id)
	return ((type(id) == "string") and (id ~= ""))
end
local isValidId = eatglobal.misc.isValidId

------------------------------------------------------------------------------------
--	eatglobal.misc.getShortFileName
--	Gibt einen verkürzten String für z.B. Debugausgaben von "fileName" zurück.
--	Aus z.B. "E:/Program Files (x86)/Steam/steamapps/workshop/content/12345/nando_truck_set_2/res/scripts/eatglobal_v2_31.lua"
--	wird ".../nando_truck_set_2/res/scripts/eatglobal_v2_31.lua"
------------------------------------------------------------------------------------
function eatglobal.misc.getShortFileName(fileName)
	local fileTails = splitString(fileName, "/")
	local result = ""
	local oneMore = false
	local countTails = #fileTails
	local endRuns = (not fileName:find("res", 1, true)) and (countTails - 1) or 1
	--for i = countTails, 1, -1 do
	for i = countTails, endRuns, -1 do
		local s = fileTails[i]
		if (i == countTails) then
			result = s
		else
			result = string.format("%s/%s", s, result)
		end
		if oneMore then break end
		if (s == "res") then
			oneMore = true
		end
	end
	return result
end
local getShortFileName = eatglobal.misc.getShortFileName

------------------------------------------------------------------------------------
--	eatglobal.misc.getValue
--	Gibt value zurück, wenn Typ von Value == Typ von default; andernfalls wird 
--	default zurückgegeben.
------------------------------------------------------------------------------------
function eatglobal.misc.getValue(value, default)
	return (type(value) ~= type(default)) and default or value
end
local getValue = eatglobal.misc.getValue

------------------------------------------------------------------------------------
--	Sucht in 's' nach 'delimiter' und gibt die Teilstücke in einer Liste zurück.
--	Ist 'plain' == true oder nil, werden keine Pattern, wie z.B. "/n" o.ä. in 's'
--	berücksichtigt.
--	Mit einer "filterFn" kann die Ausgabe gefiltert werden. Z.B. werden mit
--	"function(tail) return global.misc.isValidId(tail) end" keine "" zurückgegeben.
--
--	Beispiele:
--	s																	delimiter						Ergebnis
--	"mods/eat1963_streets/scripts"		"/"									{"mods", "eat1963_streets", "scripts"}
--	"mods/eat1963_streets/scripts/"		"/"									{"mods", "eat1963_streets", "scripts", ""}
--	"mods/eat1963_streets/scripts"		"eat1963_streets"		{"mods/", "/scripts/"}
--	"mods/eat1963_streets/scripts"		"mods"							{"", "/eat1963_streets/scripts"}
--	"mods/eat1963_streets/scripts"		"s/"								{"mod", "eat1963_street", "scripts"}
--	"1.14"														"."									{"1", "4"}
--	".1.14."													"."									{"", "1", "4", ""}
--
--	splitString wird weiter unten global gespeichert und in 'eat_string.lua'
--	global zur Verfügung gestellt. Die Ergebnisse werden als eatSetContainer
--	zurückgegeben um die entsrechenden Funktionalitäten (z.B. map, _pairs, etc.)
--	nutzen zu können.
------------------------------------------------------------------------------------
splitString = function(s, delimiter, plain, filterFn)
	local result = {}
	if (type(s) == "string") then
		filterFn = buildFilterFn(filterFn)
		local tail
		plain = getValue(plain, true)
		local from  = 1
		local delimiter_from, delimiter_to = s:find(delimiter, from, plain)
		while delimiter_from do
			tail = s:sub(from, delimiter_from - 1)
			if filterFn(tail) then
				table.insert(result, tail)
			end
			from  = delimiter_to + 1
			delimiter_from, delimiter_to = s:find(delimiter, from, plain)
		end
		tail = s:sub(from)
		if filterFn(tail) then
			table.insert(result, tail)
		end
	end
	
	return result
end

------------------------------------------------------------------------------------
--	vesriontools
--	Muss ZWINGEND in eatglobal implementiert sein, da Aufrufe in Modulen direkt beim Laden.
--	Auf keinen Fall in ein Modul auslagern. Getestet und gescheitert.
------------------------------------------------------------------------------------
local eatVersion
------------------------------------------------------------------------------------
--	Erzeugt aus 'version' eine "einfache" Version, z.B. {1, 14}.
--	version kann
--		eine Zeichenkette sein, wie z.B. "1.14" oder 1_14
--		eine volle Version {majorVersion, minorVersion}
--		oder eine einfache Version (wird direkt ohne Umwandlung ausgeggeben)
--	Wenn 'verion' ungültig ist Rückgabe {0, 0}
------------------------------------------------------------------------------------
local function makeSimpleVersion(version, default)
	if ((type(default) ~= "table") or (type(default[1]) ~= "number") or (type(default[2]) ~= "number")) then
		default = {0, 0}
	end
	if (type(version) == "string") then
		local v = splitString(version, ".")
		local result = {
			tonumber(v[1]) or 0,
			tonumber(v[2]) or 0,
		}
		if ((result[1] > 0) and (result[2] > 0)) then
			return result
		else
			v = splitString(version, "_")
			result = {
				tonumber(v[1]) or 0,
				tonumber(v[2]) or 0,
			}
		end
		if ((result[1] > 0) and (result[2] > 0)) then
			return result
		else
			return default
		end
	elseif (type(version) == "table") then
		if ((type(version[1]) == "number") and (type(version[2]) == "number")) then
			return version
		elseif ((type(version.majorVersion) == "number") and (type(version.minorVersion) == "number")) then
			return {
				version.majorVersion,
				version.minorVersion,
			}
		else
			return default
		end
	end
	return default
end
------------------------------------------------------------------------------------
--	Vergleicht 2 Versionen.
--	Rückgabe:
--		-1, wenn version1 < version2
--		1, wenn version1 > version2
--		0, wenn version1 == version2
--	Bedingungen 'version1' und version2' wie bei makeSimpleVersion
------------------------------------------------------------------------------------
local function compareVersion(version1, version2)
	local v1 = makeSimpleVersion(version1)
	local v2 = makeSimpleVersion(version2)
	if (v1[1] < v2[1]) then
		return -1
	elseif (v1[1] > v2[1]) then
		return 1
	else
		if (v1[2] < v2[2]) then
			return -1
		elseif (v1[2] > v2[2]) then
			return 1
		end
	end
	return 0
end
	
local mt_version = {
	------------------------------------------------------------------------------------
	--	__eq
	--	ACHTUNG!!!
	--	__eq wird (leider) nur aufgerufen, wenn auch die metatable identisch ist. Tja...
	------------------------------------------------------------------------------------
	__eq = function(self, var)
		return (compareVersion(self, var) == 0)
	end,
	__le = function(self, var)
		return (compareVersion(self, var) <= 0)
	end,
	__lt = function(self, var)
		return (compareVersion(self, var) < 0)
	end,
	__tostring = function(self)
		return string.format("%d.%d", self.majorVersion, self.minorVersion)
	end,
	__index = {
		simpleVersion = function(self)
			return {self.majorVersion, self.minorVersion}
		end,
	},
	__newindex = function(self, key, value) end,
	__metatable = false,
}

eatVersion = {
	new = function(version, default)
		local version = makeSimpleVersion(version, default)
		return setmetatable(
			{
				majorVersion = version[1],
				minorVersion = version[2],
			}, mt_version)
	end,
	isValidVersion = function(version)
		local _version = eatVersion.new(version)
		return ((_version.majorVersion > 0) and (_version.minorVersion > 0))
	end,
}

eatglobal.version = tostring(eatVersion.new(versionsData))
eatglobal.historie[string.format("eatglobal (Version %s)", eatglobal.version)] = historie

------------------------------------------------------------------------------------
--	Zentrale Variablenregistrierung mit Versionskennung
------------------------------------------------------------------------------------
local function checkVarInit()
	-- globale Daten
	if (type(eat1963_global_data) ~= "table") then
		eat1963_global_data = {}
	end
	-- varsByVersion
	if (type(eat1963_global_data.varsByVersion) ~= "table") then
		eat1963_global_data.varsByVersion = {}
	end
	return eat1963_global_data.varsByVersion
end

local function createSetVar(version)
	local _version = eatVersion.new(version)
	return function(var, key)
		local data = checkVarInit()
		local varData = eatglobal.misc.getValue(data[key], {})
		--if ((varData.var == nil) or (versiontools.compareVersion(varData.version, _version) < 0)) then
		if ((varData.var == nil) or (_version > varData.version)) then
			varData = {
				var = var,
				version = _version,
			}
			data[key] = varData
		end
		return varData.var
	end
end
setVar = createSetVar(versionsData)

getVar = function(key)
--local function getVar(key)
	return eatglobal.misc.getValue(checkVarInit()[key], {}).var
end

------------------------------------------------------------------------------------
--	Funktionen für's Logging
------------------------------------------------------------------------------------
local function logToString(value)
	local tv = type(value)
	if (tv == "nil") then
		return "nil"
	elseif ((tv == "boolean") or (tv == "number")) then
		return tostring(value)
	elseif (tv == "string") then
		return value
	elseif ((tv == "table")) then
		return (type(eatglobal_intern) == "table") and eatToString(value) or "type(value): "..tv
	elseif (tv == "function") then
		return "function(...)"
	else
		return "type(value): "..tv
	end
end
	
local function createLogFn(prefix, traceback)
	prefix = eatglobal.misc.isValidId(prefix) and prefix..": " or ""
	return function(...)
		local stack = {}
		for _, value in ipairs({prefix, ...}) do
			stack[#stack + 1] = logToString(value)
		end
		if (traceback == true) then
			print(debug.traceback(table.concat(stack), 2))
		else
			print(table.concat(stack))
		end
	end
end
local logFn = createLogFn("eatglobal")
local debugLogFn = __debug__ and logFn or function(...) end
local logFnTraceback = createLogFn("eatglobal", true)
local debugLogFnTraceback = __debug__ and logFnTraceback or function(...) end

local function createLogFnCounter(prefix, traceback)
	prefix = eatglobal.misc.isValidId(prefix) and prefix..": " or ""
	local counter = 1
	return function(...)
		local stack = {}
		for _, value in ipairs({"["..counter.."] ", prefix, ...}) do
			stack[#stack + 1] = logToString(value)
		end
		counter = counter + 1
		if (traceback == true) then
			print(debug.traceback(table.concat(stack), 2))
		else
			print(table.concat(stack))
		end
	end
end

------------------------------------------------------------------------------------
--	Registrieren einiger, globaler Funktionen
------------------------------------------------------------------------------------
setVar(splitString, "eat.intern.splitString")
setVar(eatVersion, "eat.intern.eatVersion")
setVar(createLogFn, "eat.intern.createLogFn")
setVar(createLogFnCounter, "eat.intern.createLogFnCounter")

------------------------------------------------------------------------------------
--	Pfadermittlung des Moduls, Laden der Untermodule (z.B. eatglobal.utils),
--	metatable für eatglobal
------------------------------------------------------------------------------------
local function getPath ()
	local function debugPath ()
		local info = debug.getinfo(1, "S")
		return (info ~= nil) and string.gsub(string.gsub(info.source, '\\', '/'), '@', "") or ""
	end
	moduleFile = debugPath()
	modulePath = string.match(moduleFile, '(.*/).*lua')
	eatglobal.apiPath = modulePath
	eatglobal.modPath = moduleFile:sub(1, moduleFile:find("/res/"))
end
getPath ()

--logFn("Initialization [", eatglobal.misc.getShortFileName(moduleFile), "] (Version ", eatglobal.version, ") ....")

local function doLoadFile(fileName)
	local f, err = loadfile(fileName)
	return (f ~= nil) and f() or nil, err
end

local moduleBlackList = {
	["global_intern"] = true,
}

------------------------------------------------------------------------------------
--	doRunPostLoadFn
--	
------------------------------------------------------------------------------------
local function doRunPostLoadFn(data, libraray)
	if (type(data.postLoadFn) == "function") then
		-------------------------------------------------------------------
		--	data.postLoadFn ausführen und anschliessend deaktivieren
		-------------------------------------------------------------------
		debugLogFn("... run postLoadFn on module 'eatglobal.", libraray, "' ...")
		data.postLoadFn({
			eatglobal = eatglobal,
			getVar = getVar,
			setVar = setVar,
			logFn = data.__debug__ and createLogFn("eatglobal."..libraray) or function(...) end,
			logFnTraceback = data.__debug__ and createLogFn("eatglobal."..libraray, true) or function(...) end,
		})
		data.postLoadFn = nil
		debugLogFn("... down running postLoadFn on module 'eatglobal.", libraray, "' ...")
	else
		-------------------------------------------------------------------
		--	keine postLoadFn vorhanden
		--	"alte" Methode ausführen
		-------------------------------------------------------------------
		debugLogFn("... no postLoadFn on module 'eatglobal.", libraray, "'. Put eatglobal, etc. on moduledata ...")
		data.eatglobal = eatglobal
		data.logFn = data.__debug__ and createLogFn("eatglobal."..libraray) or function(...) end
		data.logFnTraceback = data.__debug__ and createLogFn("eatglobal."..libraray, true) or function(...) end
	end
end

------------------------------------------------------------------------------------
--	modulDB
--	
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, fileName)
		local val = rawget(self, fileName)
		if (type(val) == "table") then
		 return val
		else
			local data, err = doLoadFile(fileName)
			if (type(data) == "table") then
				rawset(self, fileName, data)
				return data
			else
				logFnTraceback("Unable to load module '... ", getShortFileName(fileName), "'. Message: ", (err or ""))
				return nil		
			end
		end
	end,
	__newindex = function(self, key, value)
		logFnTraceback('modulDB does not support adding key "', key, '" or any other keys.')
	end,
	__metatable = false,
})

------------------------------------------------------------------------------------
--	doRegisterModule
--	
------------------------------------------------------------------------------------
local function doRegisterModule(data, library)
	-------------------------------------------------------------------
	--	data.registerModule ausführen wenn Module für die Bibliothek
	--	vorhanden sind
	-------------------------------------------------------------------
	if (type(data.registerModule) == "function") then
		local libModule = eatSetContainer(eatglobal.misc.getValue(apiSettings.module.librarys[library], {}))
		local modules = eatSetEx.new()
		debugLogFn("doRegisterModule: library = ", library)
		for mData in libModule:_pairs() do
			local modul = modulDB[mData.fileName]
			if modul then
				doRunPostLoadFn(modul, "module."..mData.modulName)
				modules:insert(modul)
			end
		end
		data.registerModule(modules)
	end
end

local libraryDB = setmetatable({}, {
	__index = {
		forceRegisterModule = function(self)
			for library, data in pairs(self) do
				debugLogFn("forceRegisterModule: library = ", library)
				doRegisterModule(data, library)
			end
		end,
	},
	__metatable = false,
})

setmetatable(eatglobal, {
	__index = function(self, key)
		debugLogFn("load module 'eatglobal.", key, "' ...")
		local val = libraryDB[key]
		if (type(val) == "table") then
			debugLogFn("... module 'eatglobal.", key, "' has already been loaded and is returned from internal table.")
			return moduleBlackList[key] and {} or val
		else
			local fileName = modulePath.."eatglobal/eat_"..key..".lua"
			
			debugLogFn("... module 'eatglobal.", key, "' is loaded from file [", fileName, "] ...")
			local data, err = doLoadFile(fileName)
			if (type(data) == "table") then
				libraryDB[key] = data
				--rawset(self, key, data)
				debugLogFn("... module 'eatglobal.", key, "' saved to internal table ...")
				-------------------------------------------------------------------
				--	data.postLoadFn
				-------------------------------------------------------------------
				doRunPostLoadFn(data, key)
				
				-------------------------------------------------------------------
				--	data.registerModule ausführen wenn Module für die Bibliothek
				--	vorhanden sind
				-------------------------------------------------------------------
				doRegisterModule(data, key)
				
				-------------------------------------------------------------------
				--	Historie abrufen
				-------------------------------------------------------------------
				eatglobal.historie[string.format("eatglobal.%s (Version %s)", key, data.version or "1.0")] = data.historie
				
				-- Ab Version 2.24 Ausgabe nur noch bei Debug.
				debugLogFn("loaded module 'eatglobal.", key, "' [", fileName, "] (Version ", data.version or "0.0", ")")
				return data
			else
				logFnTraceback("Unable to load module 'eatglobal.",key, "'. Message: ", (err or ""))
				return nil		
			end
		end
	end,
	__newindex = function(self, key, value)
		logFnTraceback('eatglobal does not support adding key "', key, '" or any other keys.')
	end,
	__metatable = false,
})

------------------------------------------------------------------------------------
-- Laden benötigter Bibliothek(en).
------------------------------------------------------------------------------------
eatglobal_intern = eatglobal.global_intern
fileutils = eatglobal.fileutils

------------------------------------------------------------------------------------
--	Prüfen auf veraltete eatglobal-Dateien und gegebenenfalls entfernen
--	Historie erstellen und speichern (eatglobal_historie.txt)
------------------------------------------------------------------------------------
local filesToDelete = eatSetEx.new({
	"eatglobal/eat_mathutils.lua",
	"eatglobal/module/eat_capacityhandler.lua",
	"eatglobal/module/eat_fileshandler.lua",
	"eatglobal/module/eat_materialshandler.lua",
	"eatglobal/module/eat_yearhandler.lua",
	"eatglobal/module/eat_modul_materialshandler.lua",
})

------------------------------------------------------------------------------------
-- checkEnvironment
--	
------------------------------------------------------------------------------------
local function checkEnvironment(apiPath, params)
	-------------------------------------------------------------------
	--	Used variables.
	-------------------------------------------------------------------
	local apisettingsFile = apiPath.."eatglobal_settings.lua"
	local apiFilesFile = apiPath.."eatglobal_files.lua"
	local oldApisettingsFile = apiPath.."apisettings.lua"
	local historieFile = apiPath.."eatglobal_historie.txt"
	local maxVersion = eatVersion.new(versionsData)
	local modApiSettings
	local apiFiles
	
	-------------------------------------------------------------------
	--	Tool(s).
	-------------------------------------------------------------------
	local function scanAllFiles()
		local function getVersion(fileName)
			-------------------------------------------------------------------
			--	ermitteln der Dateiversion anhand des Dateinamens
			-------------------------------------------------------------------
			--local result = versiontools.makeFullVersion(splitString(fileutils.getFileData(fileName).name, "eatglobal_v")[2])
			local result = eatVersion.new(splitString(fileutils.getFileData(fileName).name, "eatglobal_v")[2])
			if eatVersion.isValidVersion(result) then
				return result
			else
				return nil
			end
		end
		local function doScanMap(fileName)
			-------------------------------------------------------------------
			--	Mappingfunktion
			--	Bekommt bereits gefilterte Dateienamen (siehe doScanFilter(fileName))
			--	Wenn es eine veraltete eatglobal_v ist, Rückgabe Version und
			--	Dateiname, andernfalls nur Dateiname.
			-------------------------------------------------------------------
			local fileData = fileutils.getFileData(fileName)
			if (fileData.ext:lower() == "lua") then
				--	Nur *.lua berücksichtigen.
				if (fileData.name:find("eatglobal_v", 1, true) == 1) then
					local fileVersion = getVersion(fileData.name)
					if fileVersion then
						maxVersion = (fileVersion > maxVersion) and fileVersion or maxVersion
						return {
							fileName = fileName,
							fileData = fileData,
							version = fileVersion,
						}
					else
						return nil
					end
				else
					return {
						fileName = fileName,
						fileData = fileData,
					}
				end
			end
		end
		local function doScanFilter(fileName)
			-------------------------------------------------------------------
			--	Filterfunktion
			--	Wir nutzen hier eine Filtertabelle.
			--	Im Unterschied zu "normalen" Filtern müssen nicht alle sondern
			--	nur wenigstens eine der Bedingungen erfüllt sein.
			-------------------------------------------------------------------
			local fileData = fileutils.getFileData(fileName)
			local mapFilter = {
				--	Datei liegt im Api-Pfad und fängt mit "eatglobal_v" an.
				function() return ((fileData.path == apiPath) and (fileData.name:find("eatglobal_v", 1, true) == 1)) end,
				--	Datei liegt im Unterverzeichnis "eatglobal/".
				function() return string.ends(fileData.path, "eatglobal/") end,
				--	Datei liegt im Unterverzeichnis "eatglobal/module/".
				function() return string.ends(fileData.path, "eatglobal/module/") end,
			}
			for i, fun in ipairs(mapFilter) do
				if fun() then
					return true
				end
			end
			return false
		end
		------------------------------------------------------------------------------------
		--	
		------------------------------------------------------------------------------------
		if (os.getenv("OS") == "Windows_NT") then
			return fileutils.mapFiles(apiPath, doScanMap, doScanFilter, true)
		else
			local result = eatSetContainer(fileutils.readData(apiFilesFile)):map(function(value) return apiPath..value end)
			return result:map(doScanMap)
		end
	end
	
	-------------------------------------------------------------------
	--	Hier geht's los. :)
	-------------------------------------------------------------------

	------------------------------------------------------------------------------------
	-- Laden und evtl. Bereinigen der API-Einstellungen.
	------------------------------------------------------------------------------------
	if fileutils.fileExists(oldApisettingsFile) then
		modApiSettings = fileutils.readData(oldApisettingsFile)
		fileutils.deleteFile(oldApisettingsFile)
		fileutils.saveData(modApiSettings, apisettingsFile)
	else
		modApiSettings = fileutils.readData(apisettingsFile)
	end

	------------------------------------------------------------------------------------
	--	evtl. "params" auswerten
	------------------------------------------------------------------------------------
	local myModData
	local forceClean = (type(modApiSettings.librarys) ~= "table")
	debugLogFn("forceClean = ", forceClean)
	if (type(params) == "table") then
		local modVersion = eatVersion.new(params.modVersion, {1, 1})
		myModData = {
			path = params.modPath,
			version = modVersion,
			modId = getCurrentModId(),
		}
		debugLogFn("modVersion = ", modVersion)
		
		if (modVersion > getValue(modApiSettings.modData, {}).version) then
			forceClean = true
		end
	end
	
	------------------------------------------------------------------------------------
	--	Wir prüfen, ob die gespeicherten Module auch tatsächlich vorhanden sind.
	--	Falls nicht einen Neuscan erzwingen.
	--	Der Fall kann auftreten, wenn der Mod in ein anderes Verzeichnis verschoben
	--	wird oder auch, wenn auf steam veröffentlicht wird. Beim Verschieben stimmen
	--	die Pfadangaben nicht mehr. Aber auch beim Veröffentlichen auf steam wird eine
	--	eatglobal_settings.lua angelegt, welche beim User nicht mehr stimmt.
	------------------------------------------------------------------------------------
	if (not forceClean) then
		if (type(modApiSettings.module) == "table") then
			if (#modApiSettings.module == 0) then
				forceClean = true
			else
				for i, modul in ipairs(modApiSettings.module) do
					if (not fileutils.fileExists(modul.fileName)) then
						forceClean = true
						break
					end
				end
			end
		else
			forceClean = true
		end
	end
	
	modApiSettings.cleanOldApi = eatVersion.new(modApiSettings.cleanOldApi)
	
	if (forceClean or (modApiSettings.cleanOldApi < versionsData)) then
		------------------------------------------------------------------------------------
		--	Bereinigung der modApiSettings von evtl. veralteten Einträgen.
		--	Zurücksetzen der gescannten Dateieinträge.
		------------------------------------------------------------------------------------
		modApiSettings = {
			modData = myModData,
		}
		------------------------------------------------------------------------------------
		local function delFilesCompare(fileName1, fileName2)
			return (fileName1 == fileName2)
		end
		local delFiles = eatSetEx.new()
		local allFiles = scanAllFiles()
		debugLogFn("allFiles = ", allFiles)
		------------------------------------------------------------------------------------
		--	Process files to be deleted (from table "filesToDelete").
		------------------------------------------------------------------------------------
		for fileName in filesToDelete:_pairs() do
			fileName = apiPath..fileName
			if fileutils.fileExists(fileName) then
				delFiles:insert(fileName)
			end
		end
		
		local function mapFileNameFn(fData, params)
			return fData.fileName
		end
		------------------------------------------------------------------------------------
		--	Alte API-Dateien zur Löschliste hinzufügen.
		------------------------------------------------------------------------------------
		local oldApiFiles = {
			function(fData) return ((fData.fileData.path == apiPath) and (type(fData.version) == "table")) end,
			function(fData) return (fData.version < maxVersion) end,
		}
		delFiles:append(allFiles:map(mapFileNameFn, nil, nil, oldApiFiles))
		
		------------------------------------------------------------------------------------
		--	Alte Bibliotheks-Dateien zur Löschliste hinzufügen.
		------------------------------------------------------------------------------------
		local nonEatFiles = {
			orFunctions = {
				function(fData) return (string.ends(fData.fileData.path, "eatglobal/") and (fData.fileData.name:find("eat_", 1) ~= 1)) end,
				function(fData) return (string.ends(fData.fileData.path, "eatglobal/module/") and (fData.fileData.name:find("eat_modul_", 1) ~= 1)) end,
			},
		}
		delFiles:append(allFiles:map(mapFileNameFn, nil, nil, nonEatFiles))
		debugLogFn("delFiles = ", delFiles)
		
		------------------------------------------------------------------------------------
		--	Liste aller verfügbaren und gültigen Bibliotheks-Dateien in
		--	modApiSettings ablegen.
		------------------------------------------------------------------------------------
		local libraryFiles = {
			function(fData) return string.ends(fData.fileData.path, "eatglobal/") end,
			function(fData) return (fData.version == nil) end,
			function(fData) return (fData.fileData.name:find("eat_", 1) == 1) end,
			function(fData) return not delFiles:contains(fData.fileName, delFilesCompare) end
		}
		modApiSettings.librarys = allFiles:filter(libraryFiles)
		modApiSettings.librarys:forEach(function(fData) fData.modulName = fData.fileData.name:gsub("eat_", "") end)
		
		apiFiles = modApiSettings.librarys:map(function(fData) return "eatglobal/"..fData.fileData.name.."."..fData.fileData.ext end)
		--debugLogFn("modApiSettings.librarys = ", modApiSettings.librarys)
		
		------------------------------------------------------------------------------------
		--	Liste aller verfügbaren und gültigen Modul-Dateien in
		--	modApiSettings ablegen.
		------------------------------------------------------------------------------------
		local moduleFiles = {
			function(fData) return string.ends(fData.fileData.path, "eatglobal/module/") end,
			function(fData) return (fData.version == nil) end,
			function(fData) return (fData.fileData.name:find("eat_modul_", 1) == 1) end,
			function(fData) return not delFiles:contains(fData.fileName, delFilesCompare) end,
			function(fData)
				local data, err = fileutils.readData(fData.fileName)
				if (data and (type(data.metaData) == "table") and eatglobal.misc.isValidId(data.metaData.library)) then
					fData.library = data.metaData.library
					fData.name = data.metaData.name
					fData.modulName = fData.fileData.name:gsub("eat_modul_", "")
					return true
				else
					logFnTraceback("Unable to load module '... ", getShortFileName(fData.fileName), "'. Message: ", (err or ""))
					return false
				end
			end,
		}
		modApiSettings.module = allFiles:filter(moduleFiles)
		apiFiles:append(modApiSettings.module:map(function(fData) return "eatglobal/module/"..fData.fileData.name.."."..fData.fileData.ext end))
		--logFn("modApiSettings.module = ", modApiSettings.module)

		------------------------------------------------------------------------------------
		--	Alle zu löschenden Dateien löschen.
		------------------------------------------------------------------------------------
		if (delFiles:count() > 0) then
			fileutils.deleteFiles(delFiles)
			for fileName in delFiles:_pairs() do
				logFn("Delete old file: ....", getShortFileName(fileName))
			end
		end
		
		------------------------------------------------------------------------------------
		--	History ermitteln.
		------------------------------------------------------------------------------------
		for fData in modApiSettings.librarys:_pairs() do
			--	Jede Bibliothek wird einmal geladen. So wird die Historie aktualisiert.
			local lib = eatglobal[fData.modulName]
			--	Nebenbei ermitteln wir ob die Bibliothek Module unterstützt.
			--	Die Info ist vor allem nützlich für's Modden.
			fData.moduleSuitable = (type(lib.registerModule) == "function")
		end
		local s = ""
		for k, v in eatTableContainer(eatglobal.historie):pairs(function(a, b) return a < b end) do
			s = string.format("%s[%s]\n%s\n", s, k, v)
		end
		fileutils.saveData(s, historieFile, "TEXT")
		logFn("History created and saved in file ....", getShortFileName(historieFile), ".")
		
		------------------------------------------------------------------------------------
		--	Änderungen abspeichern.
		------------------------------------------------------------------------------------
		modApiSettings.cleanOldApi = maxVersion
		fileutils.saveData(modApiSettings, apisettingsFile)
		fileutils.saveData(apiFiles, apiFilesFile)
	end
	
	------------------------------------------------------------------------------------
	--	Moduldaten ermitteln und in apiSettings eintragen.
	------------------------------------------------------------------------------------
	for fData in eatSetContainer(modApiSettings.module):_pairs() do
		debugLogFn("fData = ", fData)
		if (type(apiSettings.module.librarys[fData.library]) ~= "table") then
			apiSettings.module.librarys[fData.library] = eatSetEx.new()
		end
		if not apiSettings.module.librarys[fData.library]:insertUnique(
			{
				modulName = fData.modulName,
				fileName = fData.fileName,
			},
			function(fData1, fData2) return (fData1.modulName == fData2.modulName) end
		)
		then
			debugLogFn(".... ", getShortFileName(fData.fileName), " bereits vorhanden?!")
		end
	end
	debugLogFn("apiSettings = ", apiSettings)
end
checkEnvironment(modulePath)

------------------------------------------------------------------------------------
-- eatglobal.forceCleanApi
--	
------------------------------------------------------------------------------------
eatglobal.forceCleanApi = function(modVersion)
	local modPath = fileutils.getModPath(1)
	checkEnvironment(
		modPath.."res/scripts/",
		{
			modPath = modPath,
			modVersion = modVersion,
		}
	)
	libraryDB:forceRegisterModule()
end

------------------------------------------------------------------------------------
-- eatglobal.loadDataFile
--	
------------------------------------------------------------------------------------
eatglobal.loadDataFile = function(fileName)
	if isValidId(fileName) then
		return fileutils.readData(fileutils.getModPath(1)..fileName)
	elseif (type(fileName) == "table") then
		local result = eatSetEx.new()
		local modPath = fileutils.getModPath(1)
		for fName in eatSetContainer(fileName):_pairs() do
			result:insert(fileutils.readData(modPath..fName))
		end
		return result
	else
		return {}
	end
end

------------------------------------------------------------------------------------
-- eatglobal.misc.createLogFn
--	
------------------------------------------------------------------------------------
eatglobal.misc.createLogFn = function(counter, traceback)
	if (not counter) then
		return createLogFn(".../"..getShortFileName(fileutils.getModulFileName(1)), traceback)
	else
		return createLogFnCounter(".../"..getShortFileName(fileutils.getModulFileName(1)), traceback)
	end
	
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------

logFn("Initialization [....", getShortFileName(moduleFile), "] (Version ", eatglobal.version, ") successfully completed: ", eatglobal.math.round(((os.clock() - startInit) * 1000), 5), " ms.")

return eatglobal